//========================================================= file = main.c =====
//=  Main program for IYER PPS (centralized) with FCFS-OQ switch planes       =
//=============================================================================
//=  Notes:  1) Simulation output is written to created file OUTFILE defined  =
//=             in common.h                                                   =
//=---------------------------------------------------------------------------=
//=  Build:  Build as a CSIM application linking to appropriate CSIM libs.    =
//=          Component files are:                                             =
//=            - common.h, main.c, demux.c, inport.c, outport.c, mux.c, and   =
//=              mux_sel.c                                                    =
//=---------------------------------------------------------------------------=
//=  History: AAS (04/01/03) - Genesis                                        =
//=============================================================================
#define MAIN                    // This is main (used in common.h)

//----- Include files ---------------------------------------------------------
#include <stdio.h>              // Needed for printf() and fprintf()
#include <stdlib.h>             // Needed for exit()
#include <math.h>               // Needed for sqrt()
#include "csim.h"               // Needed for CSIM
#include "common.h"             // Needed for common stuff

//=============================================================================
//==  Main program where sim() is the first process for a CSIM simulation    ==
//=============================================================================
void sim(void)
{
  FILE     *fp;                 // File pointer for the output file
  int      run_num;             // Simulation run number
  double   global_mean;         // Overall (global) mean delay
  double   global_std_dev;      // Overall (global) std dev delay
  int      i,j;                 // Generic loop counters

  // Set model name for CSIM
  set_model_name("IYER PPS (centralized) switch simulation model");

  // Increase maximum number of allowed processes, facilities, and servers
  max_facilities(10000);
  max_servers(100000);
  max_processes(1000000);
  max_events(1000000);
  max_messages(1000000);
  max_mailboxes(10000); 

  // Open output file for writing (delete and re-write if file already exists)
  fp = fopen(OUT_FILE, "w");
  if (fp == NULL)
  {
    printf("  >>> ERROR in creating output file '%s' \n", OUT_FILE);    
    exit(1);                
  }

  // *** Main simulation loop ***
  for (run_num=0; run_num<1; run_num++)
  {
    create("sim");

    // Initialize run parameters
    Index = 1;
    Count = Reorder = 0;
    for(i=0;i<FLOW_CELLS;i++)
      Cell_served[i] = FALSE;
    for (i=0; i<N; i++)
    {
      Outport_count[i] = 0;
      Freeslot[i] = 0;

      for (j=0; j<N; j++)
      {
        Out_seq[i][j] = 0;
      }
    }
    for (i=0; i<K; i++)
    {
      for(j=0; j<N; j++)
      {
        Ail[i][j]= TRUE;
        Ail_freetime[i][j] = 0.0;
        Aol[i][j] = TRUE;
        Aol_freetime[i][j] = 0.0;
      }
    } 
    // Assign pointers to generators, demux, mux, inports, and outports
    for (i=0; i<N; i++)
    {
      indemux[i]  = demux;
      outmux[i] = mux;
      sel[i] = mux_sel;
      if (GEN_FLAG == 1) gen[i] = generate1;
      if (GEN_FLAG == 2) gen[i] = generate2;
    }
    for (i=0; i<K; i++)
    {
      for(j=0; j<N; j++)
      {
        in[i][j]=inport;
        out[i][j]=outport;
        chan[i][j] = channel;
      }
    }
    // CSIM initializations
    for (i=0; i<K; i++)
    {
      for(j=0; j<N; j++)
      {
        Outport_internal[i][j] = facility("Server for internal outports");
        Channel_server[i][j]   = facility("Server for internal inport links");
        Q_event[i][j] = event("Queuing event at internal outports ");
        Seq_number[i][j] = mailbox("Sequence # exchange to re-assemnbler");
      }
    }
    for (i=0; i<N; i++)
    {
      Delay_table[i] = table("Cell delay table");
      Outport_external[i] = facility("Server for external outports");
    }

    // Run Parameters
    Lambda = 0.98 + (run_num*0.02);
    Sim_time = 1000000.0;

    // Kick-off the generators and sel processes
    for (i=0; i<N; i++)
    {
      gen[i](i);
      sel[i](i);
    }

    // Run until Sim_time expires while outputting status messages to stdout
    run_sim(Sim_time, run_num);

    // Compute global statistics
    global_mean = global_std_dev = 0.0;
    for (i=0; i<N; i++)
    {
      global_mean = global_mean + table_mean(Delay_table[i]) / N;
      global_std_dev = global_std_dev + sqrt(table_var(Delay_table[i])) / N;
    }

    // Output simulation configuration parameters
    fprintf(fp,"*************************************************\n");
    fprintf(fp,"== Results from IYER PPS (centralized) switch    \n");
    fprintf(fp,"*************************************************\n");
    fprintf(fp,"== Total CPU time         = %9.3f sec            \n",
      cputime());
    fprintf(fp,"== Total sim time         = %9.3f sec            \n",
      clock);
    fprintf(fp,"&** Simulation time       = %7.0f cell times     \n",
      Sim_time);
    fprintf(fp,"&** Switch size           = %d ports             \n",
      N);
    fprintf(fp,"&** Number of Planes      = %d Planes            \n",
      K);
    fprintf(fp,"&** Cell time             = %f time units        \n",
      CELL_TIME);

    if (GEN_FLAG == 1)
    {
      fprintf(fp,"&** >>> Bernoulli arrivals                     \n");
      fprintf(fp,"&** Bernoulli load      = %5.2f %%             \n",
        100.0 * Lambda);
    }
    if (GEN_FLAG == 2)
    {
      fprintf(fp,"&** >>> Interupted Bernoulli (IBP) arrivals    \n");
      fprintf(fp,"&** IBP Alpha            = %f              \n", Alpha);
      fprintf(fp,"&** IBP P                = %f              \n", P);
      fprintf(fp,"&** IBP Q                = %f              \n", Q);
      fprintf(fp,"&** IBP load             = %5.2f %%        \n",
        100.0 * Alpha * (1.0 - Q) / (2.0 - P - Q));
      fprintf(fp,"&** IBP CoV              = %f              \n",
        1.0 + Alpha * (((1.0 - P) * (P + Q)) / pow((2.0 - P - Q), 2.0) - 1.0));
    }

    fprintf(fp,"************************************************\n");
    fprintf(fp,"&** Total cells in a flow            = %d       \n",Count);
    fprintf(fp,"&** Total re-order cells in a flow   = %d       \n",Reorder);

    // Output per port statistics
    for (i=0; i<N; i++)
    {
      fprintf(fp,"%2d===============================================\n", i);
      fprintf(fp,"%2d= Outport utilization    = %5.2f %%       \n", i,
        100.0 * table_cnt(Delay_table[i]) / Sim_time);
      fprintf(fp,"%2d= Mean delay             = %f cell times  \n", i,
        table_mean(Delay_table[i]));
      fprintf(fp,"%2d= Std Dev delay          = %f cell times  \n", i,
        sqrt(table_var(Delay_table[i])));
      fprintf(fp,"%2d= Number of completions  = %5d            \n", i,
        Outport_count[i]);
      fprintf(fp,"%2d= Loss at Demultiplexor  = %d cells       \n", i,
        Demux_loss[i]);
      fprintf(fp,"%2d= Loss at Multiplexor    = %d cells       \n", i,
        Mux_loss[i]);
    }

    // Output global statistics
    fprintf(fp,"======================================================\n");
    fprintf(fp,"==  global mean delay     = %f cell times  \n",global_mean);
    fprintf(fp,"==  global std dev delay  = %f cell times  \n",global_std_dev);
    fprintf(fp,"======================================================\n");

    // Reset everything for a possible rerun
    rerun();
    reset_prob(1L);
  }

  // Close and end
  fprintf(fp,"=== END OF SIMULATION === \n");
  fclose(fp);
}
